@@ -1,4 +1,3 @@
-import math
 from typing import List, Tuple, Union
 
 import numpy as np
@@ -86,6 +85,27 @@ def spmat_interp_adjoint(
     return torch.view_as_complex(image).reshape(*output_size)
 
 
+@torch.jit.script
+def calc_split_sizes(
+    length: int,
+    num_splits: int,
+) -> List[int]:
+    """Same as np.array_split for PyTorch."""
+    # TODO: replace all calls of this function with split_tensor in PyTorch 1.8
+    size1 = length // num_splits + 1
+    num_size1 = length % num_splits
+    size2 = length // num_splits
+
+    split_sizes: List[int] = []
+    for i in range(num_splits):
+        if i < num_size1:
+            split_sizes.append(size1)
+        else:
+            split_sizes.append(size2)
+
+    return split_sizes
+
+
 @torch.jit.script
 def calc_coef_and_indices(
     tm: Tensor,
@@ -175,7 +195,7 @@ def table_interp_one_batch(
     centers = torch.floor(numpoints * table_oversamp / 2).to(dtype=int_type)
 
     # offset from k-space to first coef loc
-    base_offset = 1 + torch.floor(tm - numpoints.unsqueeze(1) / 2.0).to(dtype=int_type)
+    base_offset = 1 + torch.floor(tm - numpoints.unsqueeze(-1) / 2.0).to(dtype=int_type)
 
     # flatten image dimensions
     image = image.reshape(image.shape[0], image.shape[1], -1)
@@ -199,13 +219,79 @@ def table_interp_one_batch(
 
     # phase for fftshift
     return kdat * imag_exp(
-        torch.mv(torch.transpose(omega, 1, 0), n_shift),
+        torch.sum(omega * n_shift.unsqueeze(-1), dim=-2, keepdim=True),
         return_complex=True,
     )
 
 
 @torch.jit.script
-def table_interp_over_batches(
+def table_interp_multiple_batches(
+    image: Tensor,
+    omega: Tensor,
+    tables: List[Tensor],
+    n_shift: Tensor,
+    numpoints: Tensor,
+    table_oversamp: Tensor,
+    offsets: Tensor,
+) -> Tensor:
+    """Table interpolation with for loop over batch dimension."""
+    kdat = []
+    for (it_image, it_omega) in zip(image, omega):
+        kdat.append(
+            table_interp_one_batch(
+                it_image.unsqueeze(0),
+                it_omega,
+                tables,
+                n_shift,
+                numpoints,
+                table_oversamp,
+                offsets,
+            )
+        )
+
+    return torch.cat(kdat)
+
+
+@torch.jit.script
+def table_interp_fork_over_batchdim(
+    image: Tensor,
+    omega: Tensor,
+    tables: List[Tensor],
+    n_shift: Tensor,
+    numpoints: Tensor,
+    table_oversamp: Tensor,
+    offsets: Tensor,
+    num_forks: int,
+) -> Tensor:
+    """Table interpolation with forking over k-space."""
+
+    # indexing is worst when we have repeated indices - let's spread them out
+    split_sizes = calc_split_sizes(omega.shape[0], num_forks)
+
+    # initialize the fork processes
+    futures: List[torch.jit.Future[torch.Tensor]] = []
+    for (image_chunk, omega_chunk) in zip(
+        image.split(split_sizes), omega.split(split_sizes)
+    ):
+        futures.append(
+            torch.jit.fork(
+                table_interp_multiple_batches,
+                image_chunk,
+                omega_chunk,
+                tables,
+                n_shift,
+                numpoints,
+                table_oversamp,
+                offsets,
+            )
+        )
+
+    # collect the results
+    return torch.cat([torch.jit.wait(future) for future in futures])
+
+
+@torch.jit.script
+def table_interp_fork_over_kspace(
     image: Tensor,
     omega: Tensor,
     tables: List[Tensor],
@@ -221,6 +307,7 @@ def table_interp_over_batches(
     klength = omega.shape[1]
     omega_chunks = [omega[:, ind:klength:num_forks] for ind in range(num_forks)]
 
+    # initialize the fork processes
     futures: List[torch.jit.Future[torch.Tensor]] = []
     for omega_chunk in omega_chunks:
         futures.append(
@@ -244,6 +331,7 @@ def table_interp_over_batches(
         device=image.device,
     )
 
+    # collect the results
     for ind, future in enumerate(futures):
         kdat[:, :, ind:klength:num_forks] = torch.jit.wait(future)
 
@@ -275,33 +363,65 @@ def table_interp(
         table_oversamp: Size of table in each dimension.
         offsets: A list of offset values for interpolation.
         min_kspace_per_fork: Minimum number of k-space samples to use in each
-            process fork.
+            process fork. Only used for single trajectory on CPU.
 
     Returns:
         ``image`` interpolated to k-space locations at ``omega``.
     """
-    if image.device == torch.device("cpu"):
-        # we fork processes for indexing, so we need to do a bit of thread management
-        # for OMP to make sure we don't oversubscribe (managment not necessary for non-OMP)
-        num_threads = torch.get_num_threads()
-        factors = torch.arange(1, math.sqrt(num_threads))
-        factors = factors[torch.remainder(torch.tensor(num_threads), factors) == 0]
-        threads_per_fork = 1
-        for factor in factors:
-            # minimum k-space points per fork
-            if num_threads / factor <= omega.shape[1] / min_kspace_per_fork:
+    if omega.ndim not in (2, 3):
+        raise ValueError("omega must have 2 or 3 dimensions.")
+
+    if omega.ndim == 3:
+        if omega.shape[0] == 1:
+            omega = omega[0]  # broadcast a single traj
+
+    if omega.ndim == 3:
+        if not omega.shape[0] == image.shape[0]:
+            raise ValueError(
+                "If omega has batch dim, omega batch dimension must match image."
+            )
+
+    # we fork processes for accumulation, so we need to do a bit of thread
+    # management for OMP to make sure we don't oversubscribe (managment not
+    # necessary for non-OMP)
+    num_threads = torch.get_num_threads()
+    factors = torch.arange(1, num_threads + 1)
+    factors = factors[torch.remainder(torch.tensor(num_threads), factors) == 0]
+    threads_per_fork = num_threads  # default fallback
+
+    if omega.ndim == 3:
+        # increase number of forks as long as it's not greater than batch size
+        for factor in factors.flip(0):
+            if num_threads // factor <= omega.shape[0]:
                 threads_per_fork = int(factor)
-                break
+
+        num_forks = num_threads // threads_per_fork
+
+        if USING_OMP and image.device == torch.device("cpu"):
+            torch.set_num_threads(threads_per_fork)
+        kdat = table_interp_fork_over_batchdim(
+            image, omega, tables, n_shift, numpoints, table_oversamp, offsets, num_forks
+        )
+        if USING_OMP and image.device == torch.device("cpu"):
+            torch.set_num_threads(num_threads)
+    elif image.device == torch.device("cpu"):
+        # determine number of process forks while keeping a minimum amount of
+        # k-space per fork
+        for factor in factors.flip(0):
+            if omega.shape[1] / (num_threads // factor) >= min_kspace_per_fork:
+                threads_per_fork = int(factor)
+
         num_forks = num_threads // threads_per_fork
 
         if USING_OMP:
             torch.set_num_threads(threads_per_fork)
-        kdat = table_interp_over_batches(
+        kdat = table_interp_fork_over_kspace(
             image, omega, tables, n_shift, numpoints, table_oversamp, offsets, num_forks
         )
         if USING_OMP:
             torch.set_num_threads(num_threads)
     else:
+        # no forking for batchless omega on GPU
         kdat = table_interp_one_batch(
             image, omega, tables, n_shift, numpoints, table_oversamp, offsets
         )
@@ -309,53 +429,194 @@ def table_interp(
     return kdat
 
 
+@torch.jit.script
 def accum_tensor_index_add(image: Tensor, arr_ind: Tensor, data: Tensor) -> Tensor:
     """We fork this function for the adjoint accumulation."""
-    return image.index_add_(0, arr_ind, data)
+    if arr_ind.ndim == 2:
+        for (image_batch, arr_ind_batch, data_batch) in zip(image, arr_ind, data):
+            for (image_coil, data_coil) in zip(image_batch, data_batch):
+                image_coil.index_add_(0, arr_ind_batch, data_coil)
+    else:
+        for (image_it, data_it) in zip(image, data):
+            image_it.index_add_(0, arr_ind, data_it)
 
-
-def accum_tensor_index_put(image: Tensor, arr_ind: Tensor, data: Tensor) -> Tensor:
-    """We fork this function for the adjoint accumulation."""
-    return image.index_put_((arr_ind,), data, accumulate=True)
+    return image
 
 
 @torch.jit.script
-def fork_and_accum(image: Tensor, arr_ind: Tensor, data: Tensor, num_forks: int):
+def accum_tensor_index_put(image: Tensor, arr_ind: Tensor, data: Tensor) -> Tensor:
+    """We fork this function for the adjoint accumulation."""
+    if arr_ind.ndim == 2:
+        for (image_batch, arr_ind_batch, data_batch) in zip(image, arr_ind, data):
+            for (image_coil, data_coil) in zip(image_batch, data_batch):
+                image_coil.index_put_((arr_ind_batch,), data_coil, accumulate=True)
+    else:
+        for (image_it, data_it) in zip(image, data):
+            image_it.index_put_((arr_ind,), data_it, accumulate=True)
+
+    return image
+
+
+@torch.jit.script
+def fork_and_accum(
+    image: Tensor, arr_ind: Tensor, data: Tensor, num_forks: int
+) -> Tensor:
     """Process forking and per batch/coil accumulation function."""
     device = image.device
 
-    futures: List[torch.jit.Future[torch.Tensor]] = []
-    for batch_ind in range(image.shape[0]):
-        for coil_ind in range(image.shape[1]):
-            # if we've used all our forks, wait for one to finish and pop
-            if len(futures) == num_forks:
-                torch.jit.wait(futures[0])
-                futures.pop(0)
+    # divide the work
+    split_sizes = calc_split_sizes(image.shape[0], num_forks)
 
-            # one of these is faster on cpu, other is faster on gpu
+    # initialize the fork processes
+    futures: List[torch.jit.Future[torch.Tensor]] = []
+    if arr_ind.ndim == 2:
+        for (image_chunk, arr_ind_chunk, data_chunk) in zip(
+            image.split(split_sizes),
+            arr_ind.split(split_sizes),
+            data.split(split_sizes),
+        ):
             if device == torch.device("cpu"):
                 futures.append(
                     torch.jit.fork(
                         accum_tensor_index_put,
-                        image[batch_ind, coil_ind],
-                        arr_ind,
-                        data[batch_ind, coil_ind],
+                        image_chunk,
+                        arr_ind_chunk,
+                        data_chunk,
                     )
                 )
             else:
                 futures.append(
                     torch.jit.fork(
                         accum_tensor_index_add,
-                        image[batch_ind, coil_ind],
-                        arr_ind,
-                        data[batch_ind, coil_ind],
+                        image_chunk,
+                        arr_ind_chunk,
+                        data_chunk,
                     )
                 )
+    else:
+        for (image_chunk, data_chunk) in zip(
+            image.split(split_sizes), data.split(split_sizes)
+        ):
+            if device == torch.device("cpu"):
+                futures.append(
+                    torch.jit.fork(
+                        accum_tensor_index_put,
+                        image_chunk,
+                        arr_ind,
+                        data_chunk,
+                    )
+                )
+            else:
+                futures.append(
+                    torch.jit.fork(
+                        accum_tensor_index_add,
+                        image_chunk,
+                        arr_ind,
+                        data_chunk,
+                    )
+                )
+
+    # wait for processes to finish
+    # results in-place
     _ = [torch.jit.wait(future) for future in futures]
 
+    return image
+
 
 @torch.jit.script
-def sort_data(
+def calc_coef_and_indices_batch(
+    tm: Tensor,
+    base_offset: Tensor,
+    offset_increments: Tensor,
+    tables: List[Tensor],
+    centers: Tensor,
+    table_oversamp: Tensor,
+    grid_size: Tensor,
+    conjcoef: bool,
+) -> Tuple[Tensor, Tensor]:
+    """For loop coef calculation over batch dim."""
+    coef = []
+    arr_ind = []
+    for (tm_it, base_offset_it) in zip(tm, base_offset):
+        coef_it, arr_ind_it = calc_coef_and_indices(
+            tm=tm_it,
+            base_offset=base_offset_it,
+            offset_increments=offset_increments,
+            tables=tables,
+            centers=centers,
+            table_oversamp=table_oversamp,
+            grid_size=grid_size,
+            conjcoef=conjcoef,
+        )
+
+        coef.append(coef_it)
+        arr_ind.append(arr_ind_it)
+
+    return (torch.stack(coef), torch.stack(arr_ind))
+
+
+@torch.jit.script
+def calc_coef_and_indices_fork_over_batches(
+    tm: Tensor,
+    base_offset: Tensor,
+    offset_increments: Tensor,
+    tables: List[Tensor],
+    centers: Tensor,
+    table_oversamp: Tensor,
+    grid_size: Tensor,
+    conjcoef: bool,
+    num_forks: int,
+) -> Tuple[Tensor, Tensor]:
+    """Split work across batchdim, fork processes."""
+    if tm.ndim == 3:
+        if tm.shape[0] == 1:
+            tm = tm[0]
+
+    if tm.ndim == 2:
+        coef, arr_ind = calc_coef_and_indices(
+            tm=tm,
+            base_offset=base_offset,
+            offset_increments=offset_increments,
+            tables=tables,
+            centers=centers,
+            table_oversamp=table_oversamp,
+            grid_size=grid_size,
+            conjcoef=conjcoef,
+        )
+    else:
+        # divide the work
+        split_sizes = calc_split_sizes(tm.shape[0], num_forks)
+
+        # initialize the fork processes
+        futures: List[torch.jit.Future[Tuple[Tensor, Tensor]]] = []
+        for (tm_chunk, base_offset_chunk) in zip(
+            tm.split(split_sizes),
+            base_offset.split(split_sizes),
+        ):
+            futures.append(
+                torch.jit.fork(
+                    calc_coef_and_indices_batch,
+                    tm_chunk,
+                    base_offset_chunk,
+                    offset_increments,
+                    tables,
+                    centers,
+                    table_oversamp,
+                    grid_size,
+                    conjcoef,
+                )
+            )
+
+        # collect the results
+        results = [torch.jit.wait(future) for future in futures]
+        coef = torch.cat([result[0] for result in results])
+        arr_ind = torch.cat([result[1] for result in results])
+
+    return coef, arr_ind
+
+
+@torch.jit.script
+def sort_one_batch(
     tm: Tensor, omega: Tensor, data: Tensor, grid_size: Tensor
 ) -> Tuple[Tensor, Tensor, Tensor]:
     """Sort input tensors by ordered values of tm."""
@@ -368,6 +629,28 @@ def sort_data(
     return tm[:, indices], omega[:, indices], data[:, :, indices]
 
 
+@torch.jit.script
+def sort_data(
+    tm: Tensor, omega: Tensor, data: Tensor, grid_size: Tensor
+) -> Tuple[Tensor, Tensor, Tensor]:
+    """Sort input tensors by ordered values of tm."""
+    if omega.ndim == 3:
+        # loop over batch dimension to get sorted k-space
+        results: List[Tuple[Tensor, Tensor, Tensor]] = []
+        for (tm_it, omega_it, data_it) in zip(tm, omega, data):
+            results.append(
+                sort_one_batch(tm_it, omega_it, data_it.unsqueeze(0), grid_size)
+            )
+
+        tm_ret = torch.stack([result[0] for result in results])
+        omega_ret = torch.stack([result[1] for result in results])
+        data_ret = torch.cat([result[2] for result in results])
+    else:
+        tm_ret, omega_ret, data_ret = sort_one_batch(tm, omega, data, grid_size)
+
+    return tm_ret, omega_ret, data_ret
+
+
 def table_interp_adjoint(
     data: Tensor,
     omega: Tensor,
@@ -392,26 +675,47 @@ def table_interp_adjoint(
         numpoints: Number of neighbors in each dimension.
         table_oversamp: Size of table in each dimension.
         offsets: A list of offset values for interpolation.
-        min_kspace_per_fork: Minimum number of k-space samples to use in each
-            process fork.
+        grid_size: Size of grid to interpolate to.
 
     Returns:
         ``data`` interpolated to gridded locations.
     """
+    if omega.ndim not in (2, 3):
+        raise ValueError("omega must have 2 or 3 dimensions.")
+
+    if omega.ndim == 3:
+        if omega.shape[0] == 1:
+            omega = omega[0]  # broadcast a single traj
+
+    if omega.ndim == 3:
+        if not omega.shape[0] == data.shape[0]:
+            raise ValueError(
+                "If omega has batch dim, omega batch dimension must match data."
+            )
+
     dtype = data.dtype
     device = data.device
     int_type = torch.long
 
-    # we fork processes for accumulation, so we need to do a bit of thread management
-    # for OMP to make sure we don't oversubscribe (managment not necessary for non-OMP)
+    # we fork processes for accumulation, so we need to do a bit of thread
+    # management for OMP to make sure we don't oversubscribe (managment not
+    # necessary for non-OMP)
     num_threads = torch.get_num_threads()
-    factors = torch.arange(1, math.sqrt(num_threads)).flip(0)
+    factors = torch.arange(1, num_threads + 1)
     factors = factors[torch.remainder(torch.tensor(num_threads), factors) == 0]
-    threads_per_fork = 1
-    for factor in factors:
-        if factor <= num_threads / (data.shape[0] * data.shape[1]):
-            threads_per_fork = int(factor)
-            break
+    threads_per_fork = num_threads  # default fallback
+
+    if omega.ndim == 3:
+        # increase number of forks as long as it's not greater than batch size
+        for factor in factors.flip(0):
+            if num_threads // factor <= omega.shape[0]:
+                threads_per_fork = int(factor)
+    else:
+        # increase forks as long as it's less/eq than batch * coildim
+        for factor in factors.flip(0):
+            if num_threads // factor <= data.shape[0] * data.shape[1]:
+                threads_per_fork = int(factor)
+
     num_forks = num_threads // threads_per_fork
 
     # calculate output size
@@ -428,7 +732,7 @@ def table_interp_adjoint(
     centers = torch.floor(numpoints * table_oversamp / 2).to(dtype=int_type)
 
     # offset from k-space to first coef loc
-    base_offset = 1 + torch.floor(tm - numpoints.unsqueeze(1) / 2.0).to(dtype=int_type)
+    base_offset = 1 + torch.floor(tm - numpoints.unsqueeze(-1) / 2.0).to(dtype=int_type)
 
     # initialized flattened image
     image = torch.zeros(
@@ -441,7 +745,7 @@ def table_interp_adjoint(
     data = (
         data
         * imag_exp(
-            torch.mv(torch.transpose(omega, 1, 0), n_shift),
+            torch.sum(omega * n_shift.unsqueeze(-1), dim=-2, keepdim=True),
             return_complex=True,
         ).conj()
     )
@@ -453,7 +757,9 @@ def table_interp_adjoint(
 
     # loop over offsets and take advantage of broadcasting
     for offset in offsets:
-        coef, arr_ind = calc_coef_and_indices(
+        if USING_OMP and device == torch.device("cpu") and tm.ndim == 3:
+            torch.set_num_threads(threads_per_fork)
+        coef, arr_ind = calc_coef_and_indices_fork_over_batches(
             tm=tm,
             base_offset=base_offset,
             offset_increments=offset,
@@ -462,9 +768,16 @@ def table_interp_adjoint(
             table_oversamp=table_oversamp,
             grid_size=grid_size,
             conjcoef=True,
+            num_forks=num_forks,
         )
+        if USING_OMP and device == torch.device("cpu") and tm.ndim == 3:
+            torch.set_num_threads(num_threads)
+
+        # multiply coefs to data
+        if coef.ndim == 2:
+            coef = coef.unsqueeze(1)
+            assert coef.ndim == data.ndim
 
-        # we have to fork this multiply ourselves
         tmp = coef * data
 
         if not device == torch.device("cpu"):
@@ -473,7 +786,25 @@ def table_interp_adjoint(
         if USING_OMP and device == torch.device("cpu"):
             torch.set_num_threads(threads_per_fork)
         # this is a much faster way of doing index accumulation
-        fork_and_accum(image, arr_ind, tmp, num_forks)
+        if arr_ind.ndim == 1:
+            # fork over coils and batches
+            if device == torch.device("cpu"):
+                image = fork_and_accum(
+                    image.view(data.shape[0] * data.shape[1], output_prod),
+                    arr_ind,
+                    tmp.view(data.shape[0] * data.shape[1], -1),
+                    num_forks,
+                ).view(data.shape[0], data.shape[1], output_prod)
+            else:
+                image = fork_and_accum(
+                    image.view(data.shape[0] * data.shape[1], output_prod, 2),
+                    arr_ind,
+                    tmp.view(data.shape[0] * data.shape[1], -1, 2),
+                    num_forks,
+                ).view(data.shape[0], data.shape[1], output_prod, 2)
+        else:
+            # fork just over batches
+            image = fork_and_accum(image, arr_ind, tmp, num_forks)
         if USING_OMP and device == torch.device("cpu"):
             torch.set_num_threads(num_threads)
 
